#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_WORD_SIZE		100
#define DEBUG_INPUT_IO		0x1
#define DEBUG_DICTIONARY_IO	0x2

int debug;
const char *usage = "Usage: spellcheck [-d debug_level] dictionary input\n";

/* Is this a character that could be part of a word? */
int is_word_char(int c) {
    return isalpha(c) || c == '\'' || c == '@';
}

/* IO: Read in dictionary by word */
void read_dictionary(char *filename)
{
    FILE *fh;
    int i = 0;
    int line_num = 1;
    int c;
    char word[MAX_WORD_SIZE];

    fh = fopen(filename, "r");
    if (!fh) {
	fprintf(stderr, "Error: could not open wordlist %s: %s\n",
		filename, strerror(errno));
	exit(1);
    }

    while ((c = fgetc(fh)) != EOF) {
	if (c == '\n') {
	    line_num++;
	    if (i == 0)
		continue;
	    word[i] = '\0';
	    i = 0;
	    if (debug & DEBUG_DICTIONARY_IO)
		printf("%s\n", word);

	    /* TODO Process word here, e.g. add_to_dictionary(word) */
	} else if (i == MAX_WORD_SIZE - 1) {
	    word[i] = 0;
	    fprintf(stderr, "Word too long at wordlist line %d: %s...\n",
		    line_num, word);
	    exit(2);
	} else {
	    if (!is_word_char(c)) {
		fprintf(stderr, "Bad character '%c' in word at line %d\n",
			c, line_num);
		exit(2);
	    }
	    word[i] = c;
	    i++;
	}
    }

    fclose(fh);
}

/* "word" should be a character buffer, where *i_ref is the number of
   valid characters in the buffer (or equivalently, the index where
   the next character would be added. If the buffer is not empty, then
   terminate the buffer, pass it to check_word, and then empty the
   buffer. */
void maybe_process_word(char *word, int *i_ref) {
    if (*i_ref != 0) {
	word[*i_ref] = '\0';
	if (debug & DEBUG_INPUT_IO)
	    printf("DBG: i=%-2d word=\"%s\"\n", *i_ref, word);
	/* TODO: process word here, e.g. check_word(word) */
	*i_ref = 0;
    }
}

/* IO: Read in input by word */
void process_document(char *filename)
{
    FILE *fh;
    int i = 0;
    int c;
    int line_num = 1;
    int check_hyphen_mode = 0;
    int seen_hyphen_newline = 0;
    char word[MAX_WORD_SIZE];

    fh = fopen(filename, "r");
    if (!fh) {
	fprintf(stderr, "Error: could not open input %s: %s\n", filename,
		strerror(errno));
	exit(1);
    }

    while ((c = fgetc(fh)) != EOF) {
	if (c == '\n')
	    line_num++;
	if (check_hyphen_mode) {
	    if (isspace(c)) {
		if (c == '\n')
		    seen_hyphen_newline = 1;
		continue;
	    } else {
		check_hyphen_mode = 0;
		if (!seen_hyphen_newline) {
		    /* If no newline, treat as two words */
		    maybe_process_word(word, &i);
		} else {
		    /* If there was a newline, ignore all the
		       whitespace */
		}
		seen_hyphen_newline = 0;
	    }
	}
	if (c == '-') {
	    /* Can't decide yet whether this is the end of a word, it
	       depends what comes next. */
	    check_hyphen_mode = 1;
	} else if (!is_word_char(c)) {
	    maybe_process_word(word, &i);
	} else if (i == MAX_WORD_SIZE - 1) {
	    word[i] = 0;
	    fprintf(stderr, "Word too long in document at line %d: %s...\n",
		    line_num, word);
	    exit(2);
	} else {
	    assert(is_word_char(c));
	    word[i] = c;
	    i++;
	}
    }
    maybe_process_word(word, &i);

    fclose(fh);
}

int main(int argc, char **argv)
{
    char *dict_fname;
    char *doc_fname;

    if (argc == 3) {
	/* No options */
	dict_fname = argv[1];
	doc_fname = argv[2];
    } else if (argc == 5) {
	/* Maybe one option and its argument, which must come first. */
	if (argv[1][0] == '-') {
	    if (argv[1][1] == 'd' && argv[1][2] == '\0') {
		char *endptr;
		debug = strtol(argv[2], &endptr, 0);
		if (debug < 0 || endptr == argv[2] || *endptr != '\0') {
		    fprintf(stderr, "%s", usage);
		    fprintf(stderr, "Argument to -d should be a "
			    "non-negative integer\n");
		    return 1;
		}
		dict_fname = argv[3];
		doc_fname = argv[4];
	    } else {
		fprintf(stderr, "%s", usage);
		fprintf(stderr, "Unrecognized option '%s'\n", argv[1]);
		return 1;
	    }
	} else {
	    fprintf(stderr, "%s", usage);
	    fprintf(stderr, "Too many non-option arguments\n");
	    return 1;
	}
    } else {
	fprintf(stderr, "%s", usage);
	fprintf(stderr, "Wrong number of arguments\n");
	return 1;
    }

    read_dictionary(dict_fname);
    process_document(doc_fname);
}
